import numpy as np
import pandas as pd
from .sgd_regression import SGDRegressor

def sigmoid(x):
    ex = np.exp(x)
    return ex / (1 + ex)

class LogisticRegressor(SGDRegressor):
    """
    A class that represents a trainable logistic regression binary 
    classifier.
    Uses average negative log likelihood as the loss function.

    Attributes:
    ----------
    thresh (default=0.5)
        The threshold at which predictions from the logistic regressor
        change from 0 to 1. If thresh=None, then the model will output
        probabilities instead of classifications.
    epochs (default=5)
        The number of passes the model makes through the dataset.
    batch_size (default=1)
        The number of observations used to calculate the gradient during
        the process of stochastic gradient descent.
    learning_rate (default=0.001)
        The rate at which the model changes during stochastic gradient
        descent.
    l2_reg_weight (default=0)
        The weight used for the L2 regularization. Use 0 for no L2
        regularization.
    seed (default=0)
        The seed used for random processes.
    
    Functions:
    ----------
    calc_gradient(self, batch_X, batch_y)
        A function that calculates the gradient of the loss function.
    sgd(self, batch_X, batch_y):
        Performs one step of stochastic gradient descent.
    train(self, X, y):
        Trains the model on a dataset and the corresponding labels.
    predict(self, X):
        Makes prediction from a dataframe of observations.
    get_weights(self):
        Returns the trained weights of the model.
    """

    def __init__(self, thresh=0.5, epochs=5, batch_size=1, learning_rate=0.001, l2_reg_weight=0, seed=0):
        """
        Initializes the LogisticRegressor.

        Parameters:
        ----------
        thresh (default=0.5)
            The threshold at which predictions from the logistic regressor 
            change from 0 to 1. If thresh=None, then the model will output
            probabilities instead of classifications.
        epochs (default=5)
            The number of passes the model makes through the dataset.
            
        batch_size (default=1)
            The number of observations used to calculate the gradient during
            the process of stochastic gradient descent.
        learning_rate (default=0.001)
            The rate at which the model changes during stochastic gradient
            descent.
        seed (default=0)
            The seed used for random processes.
        """
        self.thresh = thresh
        super().__init__(epochs, batch_size=batch_size, learning_rate=learning_rate, l2_reg_weight=l2_reg_weight, seed=seed)

    def calc_gradient(self, batch_X, batch_y):
        """
        A function to calculate the gradient of the loss function.
        
        Parameters:
        ----------
        batch_X 
            A dataframe of a batch of observations without labels
        batch_y
            A dataframe (or series) of the labels of those observations.
        
        Returns:
        ----------
        gradient
            The calculated gradient.
        """
        gradient = 0
        for i, x in batch_X.iterrows():
            gradient += (sigmoid(x @ self.theta) - batch_y[i]) * x
        gradient /= batch_X.shape[0]
        return gradient

    def train(self, X, y):
        """
        Trains the LogisticRegressor using stochastic gradient descent and the
        initialized hyperparameters.

        Parameters:
        ----------
        X
            A dataframe containing rows representing observations 
            without a label column. Strictly numeric.
        y
            A dataframe (or series) containing the labels for each
            observation in X.
        """
        super().train(X, y)

    def predict(self, X):
        """
        Makes predictions from a dataframe of observations.

        Parameters:
        ----------
        X
            A dataframe (or series) containing observations.

        Returns:
        ----------
        y
            The predicted labels of the observations.
        """
        ones = pd.DataFrame(np.ones((X.shape[0], 1)))
        X_ones = pd.concat((ones, X), axis=1)
        preds = sigmoid(X_ones @ self.theta)
        if self.thresh != None:
            preds = preds >= self.thresh
        return preds


import numpy as np
import pandas as pd
from sqlearn import LinearRegressor
from sqlearn import LogisticRegressor

"""
This is a demo file that shows how square-learn models are used.
"""

def main():
    # Processing the example data (pandas)

    train_file = "examples/data/heart_test.tsv"
    test_file = "examples/data/heart_train.tsv"
    train_data = pd.read_csv(train_file, sep="\t", header=0)
    test_data = pd.read_csv(test_file, sep="\t", header=0)

    train_X = train_data.drop("heart_disease", axis=1)
    train_y = train_data["heart_disease"]
    test_X = test_data.drop("heart_disease", axis=1)
    test_y = test_data["heart_disease"]

    # Evaluating LinearRegressor

    linreg = LinearRegressor(epochs=10, batch_size=5, learning_rate=0.01,
                             l2_reg_weight=0.001, seed=5)
    linreg.train(train_X, train_y)

    linreg_preds = linreg.predict(test_X)
    linreg_preds = linreg_preds > 0.5

    linreg_errors = sum(linreg_preds != test_y)
    linreg_error_rate = linreg_errors/len(test_y)
    print(f"linreg error: {linreg_error_rate}")
    print(f"weights: {linreg.get_weights()}\n")

    # Evaluating LogisticRegressor

    logreg = LogisticRegressor(thresh=0.5, epochs=10, batch_size=5, 
                               learning_rate=0.01, l2_reg_weight=0.001, seed=4)
    logreg.train(train_X, train_y)

    logreg_preds = logreg.predict(test_X)

    logreg_errors = sum(logreg_preds != test_y)
    logreg_error_rate = logreg_errors/len(test_y)
    print(f"logreg error: {logreg_error_rate}")
    print(f"weights: {logreg.get_weights()}\n")

main()


import numpy as np
import pandas as pd
from .sgd_regression import SGDRegressor

class LinearRegressor(SGDRegressor):
    """
    A class that represents a trainable linear regressor.
    Uses mean squared error as the loss function.

    Attributes:
    ----------
    epochs (default=5)
        The number of passes the model makes through the dataset.
    batch_size (default=1)
        The number of observations used to calculate the gradient during
        the process of stochastic gradient descent.
    learning_rate (default=0.001)
        The rate at which the model changes during stochastic gradient
        descent.
    l2_reg_weight (default=0)
        The weight used for the L2 regularization. Use 0 for no L2
        regularization.
    seed (default=0)
        The seed used for random processes.
    
    Functions:
    ----------
    calc_gradient(self, batch_X, batch_y)
        A function that calculates the gradient of the loss function.
    sgd(self, batch_X, batch_y):
        Performs one step of stochastic gradient descent.
    train(self, X, y):
        Trains the model on a dataset and the corresponding labels.
    predict(self, X):
        Makes prediction from a dataframe of observations.
    get_weights(self):
        Returns the trained weights of the model.
    """

    def calc_gradient(self, batch_X, batch_y):
        """
        A function to calculate the gradient of the loss function.
        
        Parameters:
        ----------
        batch_X 
            A dataframe of a batch of observations without labels
        batch_y
            A dataframe (or series) of the labels of those observations.
        
        Returns:
        ----------
        gradient
            The calculated gradient.
        """
        gradient = 0
        for i, x in batch_X.iterrows():
            gradient += ((x @ self.theta) - batch_y[i]) * x
        gradient /= batch_X.shape[0]
        return gradient

    def predict(self, X):
        """
        Makes predictions from a dataframe of observations.

        Parameters:
        ----------
        X
            A dataframe (or series) containing observations.

        Returns:
        ----------
        y
            The predicted labels of the observations.
        """
        ones = pd.DataFrame(np.ones((X.shape[0], 1)))
        X_ones = pd.concat((ones, X), axis=1)
        preds = X_ones @ self.theta
        return preds



import numpy as np
import pandas as pd

class Model:
    """
    An interface for machine learning models.

    Functions:
    ----------
    train(self, X, y):
        Trains the model on a dataset and the corresponding labels.
    
    predict(self, X):
        Makes a prediction from a given observation.
    """ 

    def train(self, X, y):
        """
        Trains the model on a dataset and the corresponding labels.

        Parameters:
        ----------
        X
            A dataframe containing rows representing observations 
            without a label column.
        y
            A dataframe (or series) containing the labels for each
            observation in X.
        """
        pass

    def predict(self, X):
        """
        Makes predictions from a dataframe of observations.

        Parameters:
        ----------
        X
            A dataframe (or series) containing observations.

        Returns:
        ----------
        y
            The predicted labels of the observations.
        """
        pass






import numpy as np
import pandas as pd
from .model_interface import Model
import random

class SGDRegressor(Model):
    """
    A parent class for a regression model that's trained using stochastic 
    gradient descent and optional L2 regularization. 
    
    The functions calc_gradient and predict are not implemented.

    Attributes:
    ----------
    epochs (default=5)
        The number of passes the model makes through the dataset.
    batch_size (default=1)
        The number of observations used to calculate the gradient during
        the process of stochastic gradient descent.
    learning_rate (default=0.001)
        The rate at which the model changes during stochastic gradient
        descent.
    l2_reg_weight (default=0)
        The weight used for the L2 regularization. Use 0 for no L2
        regularization.
    seed (default=0)
        The seed used for random processes.
    
    Functions:
    ----------
    calc_gradient(self, batch_X, batch_y) [NOT IMPLEMENTED]
        A function that calculates the gradient of the loss function.
    sgd(self, batch_X, batch_y):
        Performs one step of stochastic gradient descent.
    train(self, X, y):
        Trains the model on a dataset and the corresponding labels.
    predict(self, X): [NOT IMPLEMENTED]
        Makes prediction from a dataframe of observations.
    get_weights(self):
        Returns the trained weights of the model.
    """

    def __init__(self, epochs=5, batch_size=1, learning_rate=0.001, l2_reg_weight=0, seed=0):
        """
        Initializes the model.

        Parameters:
        ----------
        epochs (default=5)
            The number of passes the model makes through the dataset.
        batch_size (default=1)
            The number of observations used to calculate the gradient during
            the process of stochastic gradient descent.
        learning_rate (default=0.001)
            The rate at which the model changes during stochastic gradient
            descent.
        seed (default=0)
            The seed used for random processes.
        """
        self.epochs = epochs
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.l2_reg_weight = l2_reg_weight
        self.seed = seed
        random.seed(self.seed)
        self.theta = None
        self.trained = False

    def calc_gradient(self, batch_X, batch_y):
        """
        A function to calculate the gradient of the loss function.
        
        Parameters:
        ----------
        batch_X 
            A dataframe of a batch of observations without labels
        batch_y
            A dataframe (or series) of the labels of those observations.

        Returns:
        ----------
        gradient
            The calculated gradient.
        """
        pass

    def sgd(self, batch_X, batch_y):
        """
        Performs one step of stochastic gradient descent.

        Parameters:
        ----------
        batch_X
            A dataframe containing a batch of observations.
        batch_y
            A dataframe (or series) containing the labels
            for batch_X
        """
        gradient = self.calc_gradient(batch_X, batch_y)
        self.theta = self.theta - (self.learning_rate*gradient + self.l2_reg_weight*self.theta)

    def train(self, X, y):
        """
        Trains the model using stochastic gradient descent and the
        initialized hyperparameters.

        Parameters:
        ----------
        X
            A dataframe containing rows representing observations 
            without a label column. Strictly numeric.
        y
            A dataframe (or series) containing the labels for each
            observation in X.
        """
        ones = pd.DataFrame(np.ones((X.shape[0], 1)))
        ones.index = X.index
        X_ones = pd.concat((ones, X), axis=1)
        m = X_ones.shape[0]
        n = X_ones.shape[1]
        if not self.trained:
            self.theta = pd.Series(np.zeros(n))
            self.theta = self.theta.set_axis(X_ones.columns)
            self.trained = True
        for epoch in range(self.epochs):
            for batch in range(m):
                batch_i = random.sample(range(m), self.batch_size)
                batch_X = X_ones.iloc[batch_i]
                batch_y = y.iloc[batch_i]
                self.sgd(batch_X, batch_y)

    def predict(self, X):
        """
        Makes predictions from a dataframe of observations.

        Parameters:
        ----------
        X
            A dataframe (or series) containing observations.

        Returns:
        ----------
        y
            The predicted labels of the observations.
        """
        pass

    def get_weights(self):
        """
        Returns the trained weights of the SGDRegressor.
        Can also be accessed by the attribute 'theta'.

        Returns:
        ----------
        w
            The trained weights of the model.
        """
        return self.theta
